import pluggy

from pubcontrol_scopus.observe import AuthorObservatory, AuthorObservatoryABC
from pubcontrol_scopus.request import ScopusABC, Scopus

# Added 11.04.2018
# Need the DictQuery for implementation of the adapter, that translates the internal publication dict structure into
# the structure needed by pubcontrol to be inserted into the register
from jutil.processing import DictQuery

from pubcontrol_scopus.config import Config
import datetime

import logging

from pprint import pprint

from abc import ABC, abstractmethod
import json


hookimpl = pluggy.HookimplMarker('pubcontrol')


class PublicationDictAdapter(DictQuery):

    def __init__(self):
        DictQuery.__init__(self)
        self.keywords = []
        self.type = ''
        self.publication_dict = {}

    def set(self, arg):
        assert isinstance(arg[0], dict)
        self.set_query_dict(arg[0])
        self.keywords = arg[1]
        self.type = arg[2]
        self.publication_dict = {}

        return self

    def get(self):
        self.process()
        return self.publication_dict

    def process(self):
        self.publication_dict['doi'] = self.query('doi', '')
        self.publication_dict['title'] = self.query('title', '')
        self.publication_dict['description'] = self.query('description', '')
        self.publication_dict['published'] = self.query('date', datetime.datetime.now())
        self.publication_dict['authors'] = list(map(
            lambda x: {'first_name': x['first_name'], 'last_name': x['last_name']},
            self.query('authors', [])
        ))
        self.publication_dict['journal'] = {
            'name': self.query('journal', ''),
            'volume': self.query('volume', '')
        }
        self.publication_dict['tags'] = list(map(
            lambda x: {'content': x, 'description': ''},
            self.query('keywords', [])
        ))
        self.publication_dict['categories'] = list(map(
            lambda x: {'content': x, 'description': ''},
            self.keywords
        ))
        self.publication_dict['citations'] = list(map(
            lambda x: x['scopusID'],
            self.query('citedBy', [])
        ))
        self.publication_dict['origin'] = {
            'id': self.query('scopusID', 0),
            'name': Config().NAME,
            'type': self.type,
            'text': '',
            'raw': self.query('raw', '').encode('utf-8'),
            'fetched': datetime.datetime.now()
        }

    def query_exception(self, query, query_dict, default):
        return default


class BlacklistABC(ABC):

    @abstractmethod
    def add(self, item):
        pass

    @abstractmethod
    def contains(self, item):
        pass

    def __contains__(self, item):
        return self.contains(item)

    @abstractmethod
    def remove(self, item):
        pass


class SimpleBlacklist(BlacklistABC):

    def __init__(self):
        self.config = Config()
        self.path = self.config.PATH / 'blacklist'

        self.list = json.load(self.path.open('r'))
        self.count = 0

    def add(self, item):
        self.list.append(item)
        self.count += 1
        if self.count == 5:
            self.count = 0
            json.dump(self.list, self.path.open('w+'))

    def contains(self, item):
        return item in self.list

    def remove(self, item):
        self.list.remove(item)


class Plugin:
    
    def __init__(self,
                 scopus_class=Scopus,
                 observatory_class=AuthorObservatory,
                 publication_dict_adapter=PublicationDictAdapter,
                 blacklist_class=SimpleBlacklist
                 ):

        self.scopus = scopus_class()
        assert isinstance(self.scopus, ScopusABC)
        self.observatory = observatory_class()
        assert isinstance(self.observatory, AuthorObservatoryABC)
        self.blacklist = blacklist_class()
        assert isinstance(self.blacklist, BlacklistABC)
        self.adapter = publication_dict_adapter()

        self.config = Config()
        self.logger = logging.getLogger(self.config.NAME)
        self.logger.debug('[Scopus] Logger initialized')

    @hookimpl
    def fetch(self):
        """
        CHANGELOG

        Changed 11.04.2018:
        Added a try except around the fetching of the publications from the scopus object, as they raise a
        ConnectionError in case anything went wrong with the scopus database request. Skipping the publication in
        question in case a connection error occurs

        Changed 11.04.2018:
        Before yielding the publication dicts, adapting the scopus dict structure to the dict structure needed by
        pubcontrol

        Changed 12.04.2018:


        :return:
        """
        self.logger.info('[fetch] starting...')
        author_ids = self.observatory.author_ids
        self.logger.info('[fetch] publications for {} observed authors'.format(len(author_ids)))

        # Fetching all the authors from scopus
        authors = []
        for author_id in author_ids:
            # Changed 12.04.2018
            try:
                author = self.scopus.get_author(author_id)
                self.logger.info('[fetch] author {}'.format(author_id))
            except ConnectionError:
                self.logger.warning('[fetch] Connection error for author {}'.format(author_id))
                continue

            authors.append(author)

        self.logger.info('[fetch] Received {} author profiles from scopus'.format(len(authors)))
        self.logger.info('[fetch] Fetching {} publications...'.format(
            sum(map(lambda x: len(x['publications']), authors))
        ))

        # Fetching all the publications of those authors
        for author in authors:

            for publication_id in author['publications']:

                if self.blacklist.contains(publication_id):
                    self.logger.info('[fetch] skipping publication {}'.format(publication_id))
                    continue
                # Changed 11.04.2018
                # Added the try-except to detect ConnectionErrors, which tell us to skip the publication
                try:
                    publication_dict = self.scopus.get_publication(publication_id)
                    self.logger.info('[fetch] publication {}'.format(publication_id))
                except ConnectionError:
                    self.logger.warning('[fetch] connection error for publication {}'.format(publication_id))
                    self.blacklist.add(publication_id)
                    continue

                if self.observatory.check_publication(publication_dict):

                    self.logger.info('[fetch] yielding publication: {}'.format(publication_id))
                    # First yielding all the citation publications and then the actual one
                    for _publication_dict in publication_dict['citedBy']:
                        # Changed 11.04.2018
                        yield self.adapt_publication_dict(_publication_dict, 'CITATION')
                    # Changed 11.04.2018
                    yield self.adapt_publication_dict(publication_dict, 'PUBLICATION')

    def adapt_publication_dict(self, scopus_publication_dict, type_string):
        """
        CHANGELOG

        Added 11.04.2018
        A method, which will use the adapter to convert the scopus publication dict structure into the dict structure
        needed by pubcontrol create and insert the publication model into the database.

        :param scopus_publication_dict:
        :param type_string:
        :return:
        """
        keywords = self._keywords(scopus_publication_dict)
        arg = (scopus_publication_dict, keywords, type_string)
        return self.adapter.set(arg).get()

    def _keywords(self, scopus_publication_dict):
        keywords = self.observatory.author_keywords(scopus_publication_dict)
        return keywords



import pubcontrol.config

from abc import abstractmethod

from jutil.processing import DictQuery

import os
import json
import configparser

# Added 12.04.2018
# Needed to log a few things in this module for debugging, for that the plugin name from config and obviously the
# logging module are needed
import logging
from pubcontrol_scopus.config import Config

from pprint import pprint


class AuthorObservationABC:

    @property
    @abstractmethod
    def ids(self):
        pass

    @abstractmethod
    def check(self, affiliation_list):
        pass


class AuthorObservation(AuthorObservationABC):

    def __init__(self, first_name, last_name, ids, whitelist, blacklist, keywords):
        self.first_name = first_name
        self.last_name = last_name
        self.id_list = ids
        self.whitelist = whitelist
        self.blacklist = blacklist
        self.keywords = keywords

    @property
    def ids(self):
        return self.id_list

    def check(self, affiliation_list):

        # Added 11.04.2018
        # Checking the intersection of the given affiliation list with the whitelist and blacklist of affiliations
        # In case it intersects with the whitelist, its valid, for the blacklist it is invalid.
        # In case it somehow intersects with both the whitelist is favoured. That also means if it is not explicitly
        # in the whitelist it counts as invalid

        whitelist_intersect = list(set(affiliation_list) & set(self.whitelist))
        if len(whitelist_intersect) > 0:
            return True
        else:
            return False


class AuthorObservationBuilderABC:

    @abstractmethod
    def set(self, obj):
        pass

    @abstractmethod
    def get(self):
        pass


class DictAuthorObservationBuilder(AuthorObservationBuilderABC, DictQuery):

    def __init__(self):
        DictQuery.__init__(self)
        self.parameters = {}

    def set(self, obj):
        assert isinstance(obj, dict)
        self.dict = obj
        self.parameters = {}
        return self

    def get(self):
        self.process()
        author_observation = AuthorObservation(**self.parameters)
        return author_observation

    def process(self):
        self.parameters['first_name'] = self.query('first_name', '')
        self.parameters['last_name'] = self.query('last_name', '')
        self.parameters['ids'] = json.loads(self.query('ids', '[]'))
        self.parameters['keywords'] = json.loads(self.query('keywords', '[]'))
        self.parameters['blacklist'] = json.loads(self.query('scopus_blacklist', '[]'))
        self.parameters['whitelist'] = json.loads(self.query('scopus_whitelist', '[]'))

    def query_exception(self, query, query_dict, default):
        return default


class AuthorObservatoryABC:

    @abstractmethod
    def check_publication(self, obj):
        """
        CHANGELOG

        Added 11.04.2018
        A method, that checks a single publication for whether or not the whitelist blacklist condition checks out.
        Realized I needed this, when designing the main fetch method for scopus as a generator for RAM efficiency and
        a full list of publications is never generated, instead they are processed individually

        :param obj:
        :return:
        """
        pass

    @abstractmethod
    def author_keywords(self, obj):
        """
        CHANGELOG

        Added 11.04.2018
        Realized I still needed to somehow assign the author keywords to the publications.
        A method, that takes a publication dict and returns a list of keywords that was specified for each author in
        the author observation file.

        :param obj:
        :return:
        """
        pass

    @abstractmethod
    def filter_publications(self, publication_list):
        pass

    @property
    @abstractmethod
    def author_ids(self):
        pass


class PublicationAdapter(DictQuery):

    def __init__(self):
        DictQuery.__init__(self)
        self.author_dict = {}

    def set(self, obj):
        assert isinstance(obj, dict)
        self.set_query_dict(obj)
        self.author_dict = {}
        return self

    def get(self):
        self.process()
        return self.author_dict

    def process(self):
        author_list = self.query('authors', [])
        # Getting the author id and the list with affiliations from the scopus publication dict
        for author in author_list:
            self.set_query_dict(author)
            author_id = int(self.query('authorID', '0'))
            affiliation_list = int(self.query('affiliation', '0'))
            self.author_dict[author_id] = affiliation_list

    def query_exception(self, query, query_dict, default):
        return default


class AuthorObservatory(AuthorObservatoryABC):
    """
    CHANGELOG

    Changed 12.04.2018
    Added the config and the logger attributes to the class, to enable logging for debugging purposes
    """
    def __init__(self,
                 author_observation_builder_class=DictAuthorObservationBuilder,
                 publication_adapter_class=PublicationAdapter
                 ):
        # Changed 12.04.2018
        self.config = Config()
        self.logger = logging.getLogger(self.config.NAME)

        self.observation_builder = author_observation_builder_class()
        assert isinstance(self.observation_builder, AuthorObservationBuilderABC)
        self.adapter = publication_adapter_class()

        self.pubcontrol_config = pubcontrol.config.Config()
        self.path = self.pubcontrol_config.PROJECT_PATH / self.config.NAME / 'authors.ini'

        self.authors_config_parser = configparser.ConfigParser()
        self.authors_config_parser.read(str(self.path))

        self.dict = {}
        self._build_dict()

    def author_keywords(self, obj):
        """
        CHANGELOG

        Added 11.04.2018

        :param obj:
        :return:
        """
        keywords = []
        author_dict = self.adapter.set(obj).get()
        keys_observed = set(author_dict.keys()) & set(self.dict.keys())
        for author_id in keys_observed:
            keywords += self[author_id].keywords
        # Removing duplicates
        return list(set(keywords))

    def filter_publications(self, publication_list):
        """
        CHANGELOG

        Added 11.04.2018
        for each publication in the set, a author dict will be calculated, which contains the author ids of all
        the authors as keys and the affiliation lists as values. from this dict only those author ids are being
        used, that are being observed. For each of these observed ids the observation object is consulted if the
        affiliation list checks out with the white/blacklist

        Changed 11.04.2018
        After adding the 'check_publication' method, that checks single publications I moved the logic there and
        simply used the single check here for every single obj in the list.
        Slightly more inefficient, but way simpler.

        :param publication_list:
        :return:
        """
        filtered_publications = []
        for obj in publication_list:

            if self.check_publication(obj):
                filtered_publications.append(obj)

        return filtered_publications

    def check_publication(self, obj):
        author_dict = self.adapter.set(obj).get()

        keys_observed = list(set(author_dict.keys()).intersection(set(self.dict.keys())))
        for author_id in keys_observed:
            affiliation_id = author_dict[author_id]
            is_whitelist = affiliation_id in self[author_id].whitelist
            if is_whitelist:
                return True

        # Returning False in every case, where no whitelisting has caused a True return before
        return False

    @property
    def author_ids(self):
        return self.dict.keys()

    def _build_dict(self):
        """
        CHANGELOG

        Changed 12.04.2018
        previously the builder set received the raw value from the items iteration of the config parser, this however
        was a SECTION object from the configparser module not a dict object, which would have been needed for an
        unpacking operation in the builder though, so the SECTION object is being converted to a dict now before being
        passed to the builder
        :return:
        """
        for key, value in self.authors_config_parser.items():
            if key == 'DEFAULT':
                continue
            author_observation = self.observation_builder.set(dict(value)).get()
            for author_id in author_observation.ids:
                self.dict[author_id] = author_observation

    def __getitem__(self, item):
        return self.dict[item]



import os

from jutil.configuration import config_singleton

from ScopusWp.config import PROJECT_PATH

from pubcontrol.config import Config as PubcontrolConfig

import pathlib

NAME = 'scopus'
DB_NAME = '__scopus'
PATH = os.path.dirname(os.path.realpath(__file__))


class Config(PubcontrolConfig().derive(NAME)):

    NAME = NAME
    PATH = pathlib.Path(os.path.dirname(os.path.realpath(__file__)))





from pprint import pprint

import logging
import urllib.parse as urlparse
import requests
import json

from jutil.processing import DictQuery

import os
import datetime

from abc import abstractmethod

from pubcontrol_scopus.config import Config

# Changed 13.04.2018
# Ran into a problem with the database encoding and chose to simply get rid of all the unicode in the raw input of the
# scopus database
from unidecode import unidecode


class ScopusABC:

    @abstractmethod
    def get_author(self, author_id):
        pass

    @abstractmethod
    def get_publication(self, publication_id):
        pass


# Changed 11.04.2018
# In case of an Exception during the request, a ConnectionError is invoked, so that the top level knows to skip the
# the publication

class ScopusFetcher(DictQuery):

    def __init__(self, api_key, url):
        super().__init__()

        self.key = api_key
        self.base_url = url

        # Setting the header for the request to tell the scopus website, we want the returned data structure in JSON
        # format.
        self.headers = {
            'Accept': 'application/json',
            'X-ELS-APIKey': self.key
        }
        self.config = Config()
        self.logger = logging.getLogger(self.config.NAME)
        self.url = ''

        self.id = None
        # Will contain the string of the whole response json structure, so unused fields can be analysed later on
        # eventually
        self.raw = None

    def request_retrieval(self, view, url_extension, initial_query, timeout_tuple):
        """
        CHANGELOG

        Changed 12.04.2018
        After executing the inital query, the method checks if that query has succeeded, because for every correct
        response this query has to be correct in case it is not a ConnectionError will be invoked to notify the higher
        level to skip the author in the current processing

        :param view:
        :param url_extension:
        :param initial_query:
        :param timeout_tuple:
        :return:
        """
        # This dict will be converted into the URL query for the API page. It specifies, that scopus is supposed to
        # send as much data as possible with the service level of the given api key
        query = {
            'view': view
        }

        # The url to the general abstract retrieval api page of scopus
        url_extension_split = url_extension.split('/')
        api_url = os.path.join(self.base_url, url_extension_split[0], url_extension_split[1], str(self.id))
        # Encoding the dict query into the url
        query_encoded = urlparse.urlencode(query)
        self.url = '{}?{}'.format(api_url, query_encoded)

        # Sending the request, but expecting a timeout error from the connection. In this case the error will be
        # mapped onto the more general Connection error, which will be catched at the top level.
        try:
            # Setting a relatively low timeout for both (connect, read), as the network is not supposed to slow the
            # the whole program down, better to skip a publication than to loose so much time
            response = requests.get(self.url, headers=self.headers, timeout=timeout_tuple)
            self.raw = response.text

            # Setting the main query dict to the json abstract retrieval
            abstract_retrieval_dict = json.loads(response.text)
            self.set_query_dict(abstract_retrieval_dict)
            # Setting the query dict to the sub level dict specified with the initial query offset

            # Changed 12.04.2018
            # In case the initial query fails, that is an indicator for the response containing an error
            initial = self.query(initial_query, None)
            if initial is None:
                raise ConnectionError
            else:
                self.set_query_dict(initial)
        except Exception as e:
            self.logger.error('[Web] Could not retrieve for {} exception: {}'.format(self.id, str(e)))

            # Changed 11.04.2018
            # The ConnectionError has been chosen by me to signal on the top level, that a request to scopus has gone
            # wrong in some way and that the publication for that request has to be skipped or otherwise handled
            raise ConnectionError

    def request_scopus_search(self, search_query, start, increment, initial_query, timeout_tuple):
        """
        CHANGELOG

        Changed 12.04.2018
        After executing the inital query, the method checks if that query has succeeded, because for every correct
        response this query has to be correct in case it is not a ConnectionError will be invoked to notify the higher
        level to skip the author in the current processing

        :param search_query:
        :param start:
        :param increment:
        :param initial_query:
        :param timeout_tuple:
        :return:
        """
        query = {
            'query': search_query,
            'count': str(increment),
            'start': str(start)
        }

        # The general url to the scopus search api
        api_url = os.path.join(self.base_url, 'search', 'scopus')
        # Encoding the query into the url
        query_encoded = urlparse.urlencode(query)
        self.url = '{}?{}'.format(api_url, query_encoded)

        # Sending the request, but expecting a timeout error from the connection. In this case the error will be
        # mapped onto the more general Connection error, which will be catched at the top level.
        try:
            # Setting a relatively low timeout for both (connect, read), as the network is not supposed to slow the
            # the whole program down, better to skip a publication than to loose so much time
            response = requests.get(self.url, headers=self.headers, timeout=timeout_tuple)

            search_retrieval_dict = json.loads(response.text)
            self.set_query_dict(search_retrieval_dict)

            # Changed 12.04.2018
            # In case the initial query fails, that is an indicator for the response containing an error
            initial = self.query(initial_query, None)
            if initial is None:
                raise ConnectionError
            else:
                self.set_query_dict(initial)
        except Exception as e:
            self.logger.error('[Web] Could not retrieve citation search for {} exception: {}'.format(self.id, str(e)))

            # Changed 11.04.2018
            # The ConnectionError has been chosen by me to signal on the top level, that a request to scopus has gone
            # wrong in some way and that the publication for that request has to be skipped or otherwise handled
            raise ConnectionError


# Changed 11.04.2018
# Added the field 'collaboration' to the keys of the publication dict. It will contain 'none' if the publication is not
# part of a collaboration and the name of the collaboration otherwise

# Changed 11.04.2018
# Added the method 'get_keywords', which will return the publication wide keywords, that come with the scopus raw data
# also implemented, that those keywords will be assigned to the field 'keywords' of the publication dict, which
# already existed, but was unused in the 'process' method

class ScopusPublicationBuilder(DictQuery):

    DEFAULT_AUTHOR_DICT = {
        'authorID': 0,
        'first_name': '',
        'last_name': '',
        'source': '',
        'affiliation': 0
    }

    DEFAULT_PUBLICATION_DICT = {
        'scopusID': 0,
        'eid': '',
        'doi': '',
        'title': '',
        'description': '',
        'date': datetime.datetime.now(),
        'fetched': datetime.datetime.now(),
        'creator': DEFAULT_AUTHOR_DICT,
        'authors': [],
        'citedBy': [],
        'keywords': [],
        'journal': '',
        'volume': '',
        'source': '',
        'raw': '',
        'collaboration': 'none'
    }

    DATETIME_FORMAT = '%Y-%m-%d'

    def __init__(self):
        DictQuery.__init__(self)
        self.publication_dict = {}
        self.id = 0
        self.config = Config()
        self.logger = logging.getLogger(self.config.NAME)

    def set(self, obj):
        assert isinstance(obj, dict)
        self.set_query_dict(obj)
        self.id = self.get_id()
        self.publication_dict = {}

        return self

    def get(self):
        self.process()
        return self.publication_dict

    def process(self):
        self.publication_dict['raw'] = json.dumps(self.dict)
        self.publication_dict.update(self.get_coredata_dict())

        # Changed 11.04.2018
        self.publication_dict['keywords'] = self.get_keywords()
        self.publication_dict['collaboration'] = self.get_collaboration()

        authors = self.get_authors()
        self.publication_dict['authors'] = authors
        if len(authors) != 0:
            self.publication_dict['creator'] = authors[0]
        else:
            self.publication_dict['creator'] = self.get_creator()

    def get_collaboration(self):
        """
        CHANGELOG

        Changed 12.04.2018
        Now the original query dict is being saved at the bigging and set as the query dict at the end, so that if a
        collaboration dict is found, it can be set as query dict, because the call to the dict may cause an error and
        in this case the error handling of the query method is desireable

        :return:
        """
        original_dict = self.dict

        author_group_list = self.query('item/bibrecord/head/author-group', [])
        collaboration = self.DEFAULT_PUBLICATION_DICT['collaboration']
        for item in author_group_list:
            if isinstance(item, dict) and 'collaboration' in item.keys():
                self.set_query_dict(item)
                collaboration = self.query('collaboration/ce:indexed-name', 'none')
                self.logger.info('[request] publication {} collab "{}"'.format(self.id, collaboration))
        # The default value in case no specific name has been returned
        self.set_query_dict(original_dict)
        return collaboration

    def get_id(self):
        """
        Gets the scopus id from the response dict structure.
        Is a separate function as the value needs additional string processing
        :return: the int id
        """
        return int(self.query('coredata/dc:identifier', 'SCOPUS_ID:0').replace('SCOPUS_ID:', ''))

    def get_coredata_dict(self):
        """
        Returns a partial publication dict, which contains all the the coredata, which can be acquired from the
        returned json data structure, without further processing.
        Can be used to update the publication dict with
        :return: dict
        """
        if 'error' in self.query('coredata', {}):
            raise ConnectionError
        return {
            'scopusID': self.get_id(),
            'source': self.query('coredata/prism:url', self.DEFAULT_PUBLICATION_DICT['source']),
            'title': unidecode(self.query('coredata/dc:title', self.DEFAULT_PUBLICATION_DICT['title'])).replace('"', "'"),
            'description': unidecode(self.query('coredata/dc:description', self.DEFAULT_PUBLICATION_DICT['description'])),
            'journal': self.query('coredata/prism:publicationName', self.DEFAULT_PUBLICATION_DICT['journal']),
            'volume': self.query('coredata/prism:volume', self.DEFAULT_PUBLICATION_DICT['volume']),
            'doi': self.query('coredata/prism:doi', self.DEFAULT_PUBLICATION_DICT['doi']),
            'eid': self.query('coredata/eid', self.DEFAULT_PUBLICATION_DICT['eid']),
            'date': datetime.datetime.strptime(self.query(
                'coredata/prism:coverDate',
                self.DEFAULT_PUBLICATION_DICT['date'].strftime(self.DATETIME_FORMAT)
            ), self.DATETIME_FORMAT)
        }

    def get_creator(self):
        """
        CHANGELOG

        Changed 13.04.2018
        Added a unidecode for the creator name, so that the program is not running into trouble with the database
        encoding later on

        :return:
        """
        creator_full = unidecode(self.query('coredata/dc:creator', ' '))
        creator_split = creator_full.split(' ')
        creator_dict = self.DEFAULT_AUTHOR_DICT
        creator_dict.update({'last_name': creator_split[0], 'first_name': creator_split[1]})
        return creator_dict

    def get_keywords(self):
        keywords = []
        original_dict = self.dict
        for item in self.query('idxterms/mainterm', []):
            self.set_query_dict(item)
            keyword = self.query('$')
            if keyword != '':
                keywords.append(keyword)
        self.set_query_dict(original_dict)
        return list(set(keywords))

    def get_authors(self):
        """
        The list of author dicts

        CHANGELOG

        Changed 13.04.2018
        Added a unidecode for the author first name and last name, so that the program is not running into trouble with
        the database encoding later on

        Changed 14.04.2018
        Added a set, that will gather all the full names of the authors and only authors whose name are not already
        added will be appended to the list of author dicts.
        This is for the duplicate key protection of the database.

        :return: [author dicts]
        """
        author_names = {' '}
        author_list = []
        original_dict = self.dict
        for item in self.query('authors/author', default=[]):
            self.set_query_dict(item)
            # Default values for the author dict
            author_dict = self.DEFAULT_AUTHOR_DICT.copy()
            author_dict['authorID'] = int(self.query('@auid', '0'))
            first_name = unidecode(self.query('preferred-name/ce:given-name', ' '))
            last_name = unidecode(self.query('preferred-name/ce:surname', ' '))
            full_name = '{} {}'.format(first_name, last_name)

            author_dict['first_name'] = first_name
            author_dict['last_name'] = last_name
            author_dict['affiliation'] = int(self.query('affiliation/@id', '0'))

            # Only adding non duplicate entries for the first and last name to the list
            if full_name not in author_names:
                author_names.add(full_name)
                author_list.append(author_dict)

        # Changed 11.04.2018
        # Resetting the internal query dict to be the original dict, it was before working with the authors
        self.set_query_dict(original_dict)

        return author_list

    def query(self, dict_query, default=None):
        """
        CHANGELOG

        Changed 14.04.2018
        Overwrote the super method to add behaviour, that in any case if None is actually the querried value from the
        dict, the default value will be returned instead, because a NoneType should really not get into a database
        insert.

        :param dict_query:
        :param default:
        :return:
        """
        query_result = super(ScopusPublicationBuilder, self).query(dict_query, default)
        if query_result is None:
            return default
        return query_result

    def query_exception(self, query, query_dict, default):
        """
        This method will be called if a dict query to the current query dict fails, it gets passed the query for which
        it failed the query dict and the default value chosen for the query.
        In case of failure, the error is simply being logged and the default is returned

        :param query: The query to the dict, that failed
        :param query_dict: The query dict which could not handle the query
        :param default: The default value given
        :return: the default value
        """
        self.logger.warning('[Web] Could not query {} for {}'.format(
            self.id,
            query
        ))
        return default


class ScopusPublicationFetcher(ScopusFetcher):

    def __init__(self, api_key, url, publication_builder_class=ScopusPublicationBuilder):
        super().__init__(api_key, url)

        self.publication_dict = {}
        self.builder = publication_builder_class()

    def set(self, scopus_id):
        """
        Sets the new scopus id fot the publication to be fetched from the scopus database
        :param scopus_id: the int id
        :return: void
        """
        self.id = scopus_id
        return self

    def get(self):
        """
        Returns the fetched scopus publication dict
        :return:
        """
        self.process()
        return self.publication_dict

    def process(self):
        """
        This function executes the whole process of getting the data from the scopus website and constructing the
        publication dict from currently set scopus id
        :return: void
        """
        self.request_abstract_retrieval()
        self.publication_dict = self.builder.set(self.dict).get()
        # Requesting the citation search and for each item in the response also calling the builder
        self.request_citations_search()
        self.publication_dict['citedBy'] = []
        for raw_dict in self.query('entry', []):
            publication_dict = self.builder.set({'coredata': raw_dict}).get()
            self.publication_dict['citedBy'].append(publication_dict)

    def request_citations_search(self, start=0, count=200):
        """
        Sends a request to retrieve all publications, that have cited the currently processed publication to scopus and
        sets the returned json dict structure as the current query dict
        :param start: The start of the list
        :param count: The amount of publications per page
        :return: void
        """
        self.request_scopus_search(
            'refeid({})'.format(self.publication_dict['eid']),
            start,
            count,
            'search-results',
            (15, 15)
        )

    def request_abstract_retrieval(self):
        """
        Sends a abstract retrieval request for the currently set scopus id to the api sites and sets the returned
        json dict structure as the current query dict
        :return: void
        """
        self.request_retrieval(
            'FULL',
            'abstract/scopus_id',
            'abstracts-retrieval-response',
            (15, 15)
        )

    def query_exception(self, query, query_dict, default):
        """
        This method will be called if a dict query to the current query dict fails, it gets passed the query for which
        it failed the query dict and the default value chosen for the query.
        In case of failure, the error is simply being logged and the default is returned
        :param query: The query to the dict, that failed
        :param query_dict: The query dict which coul not handle the query
        :param default: The default value given
        :return: the default value
        """
        self.logger.warning('[Web] Could not query {} for {}'.format(
            query,
            self.id
        ))
        return default


class ScopusAuthorFetcher(ScopusFetcher):

    DEFAULT_AUTHOR_DICT = {
        'authorID': 0,
        'created': datetime.datetime.now(),
        'fetched': datetime.datetime.now(),
        'first_name': '',
        'last_name': '',
        'activity_range': (0, 0),
        'citedBy_count': 0,
        'document_count': 0,
        'source': '',
        'raw': '',
        'h_index': 0,
        'publications': []
    }

    def __init__(self, api_key, url, publication_intervals=25):
        super().__init__(api_key, url)

        self.author_dict = {}
        self.publication_intervals = publication_intervals

    def set(self, author_id):
        self.id = author_id

        return self

    def get(self):
        self.process()
        return self.author_dict

    def process(self):
        # Setting the current query dict to the author retrieval dict
        self.request_author_retrieval()

        self.author_dict = self.DEFAULT_AUTHOR_DICT.copy()
        self.author_dict.update(self.get_coredata_dict())
        self.author_dict['raw'] = self.raw

        self.author_dict['publications'] = self.get_publications()

    def get_publications(self):
        scopus_id_list = []
        document_count = self.author_dict['document_count']
        for i in range(-(-document_count // self.publication_intervals)):
            # Setting the search result dict as current query dict
            self.request_publication_search(i * self.publication_intervals)

            for item in self.query('entry', []):
                self.set_query_dict(item)
                scopus_id = int(self.query('dc:identifier', 'SCOPUS_ID:0').replace('SCOPUS_ID:', ''))
                if scopus_id == 0:
                    continue
                scopus_id_list.append(scopus_id)
        return scopus_id_list

    def get_coredata_dict(self):
        return {
            'authorID': self.get_id(),
            'affiliation': int(self.query('affiliation-current/@id', '0')),
            'first_name': self.query('author-profile/preferred-name/given-name', self.DEFAULT_AUTHOR_DICT['first_name']),
            'last_name': self.query('author-profile/preferred-name/given-name', self.DEFAULT_AUTHOR_DICT['last_name']),
            'citedBy_count': int(self.query('coredata/cited-by-count', self.DEFAULT_AUTHOR_DICT['citedBy_count'])),
            'document_count': int(self.query('coredata/document-count', self.DEFAULT_AUTHOR_DICT['document_count'])),
            'source': self.query('coredata/prism:url', self.DEFAULT_AUTHOR_DICT['source']),
            'h_index': int(self.query('h-index', self.DEFAULT_AUTHOR_DICT['h_index']))
        }

    def get_id(self):
        return int(self.query('coredata/dc:identifier', 'AUTHOR_ID:0').replace('AUTHOR_ID:', ''))

    def request_author_retrieval(self):
        """
        Sends a abstract retrieval request for the currently set scopus id to the api sites and sets the returned
        json dict structure as the current query dict
        :return: void
        """
        self.request_retrieval(
            'ENHANCED',
            'author/author_id',
            'author-retrieval-response/0',
            (15, 15)
        )

    def request_publication_search(self, start):
        self.request_scopus_search(
            'AU-ID({})'.format(self.id),
            start,
            self.publication_intervals,
            'search-results',
            (30, 30)
        )

    def query_exception(self, query, query_dict, default):
        self.logger.warning((
            '[Web] The query {} for author {} failed'
        ).format(query, self.id))
        return default


class Scopus(ScopusABC):

    def __init__(self,
                 publication_fetcher_class=ScopusPublicationFetcher,
                 author_fetcher_class=ScopusAuthorFetcher
                 ):
        self.config = Config()

        self.key = self.config['SCOPUS']['key']
        self.url = self.config['SCOPUS']['url']

        self.publication_fetcher = publication_fetcher_class(self.key, self.url)
        self.author_fetcher = author_fetcher_class(self.key, self.url)

    def get_author(self, author_id):
        return self.author_fetcher.set(author_id).get()

    def get_publication(self, publication_id):
        return self.publication_fetcher.set(publication_id).get()


if __name__ == '__main__':
    publication_fetcher = ScopusPublicationFetcher('3396e90e0692ebaa0496efa66d481c5c', 'https://api.elsevier.com/content')
    pd = publication_fetcher.set(1242311986).get()
    pprint(pd)